@@ -69,6 +69,8 @@ enum configuration_options {
 	OPT_PRESSURE_RANGE,
 	OPT_CALIBRATION,
 	OPT_AREA,
+	OPT_3FG_DRAG,
+	OPT_SENDEVENTS,
 };
 
 #define CONFIGURATION_OPTIONS \
@@ -91,6 +93,7 @@ enum configuration_options {
 	{ "disable-dwtp",              no_argument,       0, OPT_DWTP_DISABLE }, \
 	{ "enable-scroll-button-lock", no_argument,       0, OPT_SCROLL_BUTTON_LOCK_ENABLE }, \
 	{ "disable-scroll-button-lock",no_argument,       0, OPT_SCROLL_BUTTON_LOCK_DISABLE }, \
+	{ "enable-3fg-drag",           required_argument, 0, OPT_3FG_DRAG }, \
 	{ "set-click-method",          required_argument, 0, OPT_CLICK_METHOD }, \
 	{ "set-clickfinger-map",       required_argument, 0, OPT_CLICKFINGER_MAP }, \
 	{ "set-scroll-method",         required_argument, 0, OPT_SCROLL_METHOD }, \
@@ -98,6 +101,7 @@ enum configuration_options {
 	{ "set-profile",               required_argument, 0, OPT_PROFILE }, \
 	{ "set-tap-map",               required_argument, 0, OPT_TAP_MAP }, \
 	{ "set-speed",                 required_argument, 0, OPT_SPEED },\
+	{ "set-sendevents",            required_argument, 0, OPT_SENDEVENTS },\
 	{ "apply-to",                  required_argument, 0, OPT_APPLY_TO },\
 	{ "set-custom-points",         required_argument, 0, OPT_CUSTOM_POINTS },\
 	{ "set-custom-step",           required_argument, 0, OPT_CUSTOM_STEP },\
@@ -141,6 +145,8 @@ struct tools_options {
 	double pressure_range[2];
 	float calibration[6];
 	struct libinput_config_area_rectangle area;
+	enum libinput_config_3fg_drag_state drag_3fg;
+	enum libinput_config_send_events_mode sendevents;
 };
 
 void tools_init_options(struct tools_options *options);
